using System;
using System.Diagnostics;


/**
 * \file Position.cs
 * \brief Contains the Position class which defines a GPS position.
 *  
 * A position is a location on the planet indicated by a combination
 * of longtitude, lattitude. Longtitude and lattitude are given in
   degrees, in DDD format.

   Height is not included, this is a property of the waypoint.
 *
 * A position is always assumed to be relative to the WGS84
 * datum. Other datums needs to be handled through conversions. 
 */

/* 
 * The computations from this class are taken from the acccomp.tcl and
 * compute.tcl
 * files in gpsman. This was in turn taken from the "inverse" program
 * at ftp://www.ngs.noaa.gov/pub/pcsoft/for_inv.3d/
 */

/**
 * \class Position
 * \brief A position is a location as determined by GPS, using
 * latitude, longitude, and height.
 */

public class Position {

	/**
	 * Semimajor axis equatorial (in meters, for WGS84 ellipsoid)
	 */
	const int a = 6378137;

	/**
	 * Flattening (for WGS84 ellipsoid)
	 */
	const double f = 1 / 298.257223562;

	/**
	 * A 180th part of PI, to be used to do the calculations on
	 * degrees with. I guess we only use 180 because we have
	 * north/south and east/west.
	 */
	const double Radial = Math.PI / 180;
		
	/**
	 * The distance measured by degrees east or west from the prime meridian.
	 */
	double longitude;
	public double Longitude {
		get {
			return longitude;
		}
		set {
			longitude = value;
		}
	}

	/**
	 * The angular distance north or south from the equator measured in degrees.
	 */
	double latitude;
	public double Latitude {
		get {
			return latitude;
		}
		set {
			latitude = value;
		}
	}

	public Position () : this (0, 0) {
	}

	public Position (double latitude, double longitude) 
	{
		this.latitude = latitude;
		this.longitude = longitude;
	}


	/**
	 * Compute the distance between this Position and the
	 * OtherPosition given as argument.
	 */

	public double DistanceTo (Position OtherPosition)
	{
		Debug.Assert (OtherPosition != null, "OtherPosition == null");
		Debug.Assert (longitude != 0, "longitude == 0");
		Debug.Assert (latitude != 0, "latitude == 0");

		/*
		 * These Positions are at the same place. Height does not
		 * count in this calculation.
		 */
		if ( longitude == OtherPosition.Longitude &&
			 latitude == OtherPosition.Latitude)
			return 0;

		double lat1 = latitude * Radial;
		double long1 = longitude * Radial;
		double lat2 = OtherPosition.Latitude * Radial;
		double long2 = OtherPosition.Longitude * Radial;

		/* 
		 * The following formula is not always fully accurate. Use it
		 * anyway for now because it is more simple and because the
		 * complicated formula below does not give the right answer.
		 */
#if true
		return 1e-3 * a * Math.Acos(Math.Cos(long1 - long2) * Math.Cos(lat1) * Math.Cos(lat2) + Math.Sin(lat1) * Math.Sin(lat2));
#else
		/*
		 * This formula is quick and easy but not always correct. We
		 * only use it when we know the result will be ok. Otherwise
		 * we run a more complicated but accurate calculation.
		 */
		if ( Math.Abs(Math.Cos(lat1)) < 1e-20 | 
			 Math.Abs(Math.Cos(lat2)) < 1e-20 |
			 (latitude + OtherPosition.Latitude < 1e-4 & Math.Abs(Math.Abs(longitude - OtherPosition.Longitude) - 180) < 1e-4)) {
			return 1e-3 * a * Math.Acos(Math.Cos(long1 - long2) * Math.Cos(lat1) * Math.Cos(lat2) + Math.Sin(lat1) * Math.Sin(lat2));
		}

		/*
		 * Now we do some honest calculations to get the result
		 */
		
		double c2a = 0;
		double c;
		double e = 0;
		double sy = 0;
		double cy = 0;
		double y = 0;
		double sa;
		double cz = 0;
		double sx = 0;
		double cx = 0;

		const double eps = 5e-12;
		const double r = 1 - f;

		double tu1 = r * Math.Tan (lat1);
		double tu2 = r * Math.Tan (lat2);

		double cu1 = 1 / Math.Sqrt (tu1 * tu1 + 1);
		double cu2 = 1 / Math.Sqrt (tu2 * tu2 + 1);

		double su1 = cu1 * tu1;
		double s = cu1 * cu2;

		double baz = s * tu2;
		double faz = baz * tu1;
	   
		double x = long2 - long1;
		double d = x + 1;

		while ( Math.Abs (d - x) > eps ) {
			sx = Math.Sin (x);
			cx = Math.Cos (x);
			tu1 = cu2 * sx;
			tu2 = baz - su1 * cu2 * cx;

			sy = Math.Sqrt (tu1 * tu1 + tu2 * tu2);
			cy = s * cx + faz;
			y = Math.Atan2 (sy, cy);
			sa = s * sx / sy;
			c2a = - sa * sa + 1;
			cz = faz + faz;

			if ( cz > 0 ) {
				cz = - cz / c2a + cy;
			}

			e = cz * cz * 2 - 1;
			c = ((-3 * c2a + 4) * f + 4) * c2a * f / 16;
			d = x;
			x = ((e * cy * c + cz) * sy * c + y) * sa;
			x = (1 - c) * x * f + long2 - long1;
		}

		faz = Math.Atan2 (tu1, tu2);
		baz = Math.Atan2 (cu1 * sx, baz * cx - su1 * cu2) + Math.PI;

		x = Math.Sqrt ((1 / r / r - 1) * c2a + 1) + 1;
		x = (x - 2) / x;
		c = (x * x / 4 + 1) / (1 - x);
		d = (0.375 * x * x - 1) * x;
		x = e * cy;
		s = 1 - e - e;
		s = ((((sy * sy * 4 - 3) * s * cz * d / 6 - x) * d / 4 + cz) * sy * d * y) * c * a * r * 1e-3;

		return s;
#endif
	}

	/** 
	 * Calculate the bearing towards this point, in degrees.
	 */

	public double BearingTo (Position OtherPosition) {

		double bearing;

		double lat1 = latitude * Radial;
		double long1 = longitude * Radial;
		double lat2 = OtherPosition.Latitude * Radial;
		double long2 = OtherPosition.Longitude * Radial;

		double lat_diff = lat1 - lat2;
		double long_diff = long1 - long2;

		if ( Math.Abs(lat_diff) < 1e-20 ) {
			if ( Math.Abs(long_diff) < 1e-20) {
				bearing = 0;
			} else
				if ( long_diff < 0 ) {
					bearing = 270;
				} else {
					bearing = 90;
				}
			if ( Math.Abs(long_diff) < 1e-20) {
				if (lat_diff < 0) {
					bearing = 180;
				} else {
					bearing = 0;
				}
			}
		} else {
			bearing = Math.Round (Math.Atan2 (Math.Sin (long_diff), Math.Tan (lat2) * Math.Cos (lat1) - Math.Sin (lat1) * Math.Cos (long_diff)) / Radial);
		}
		
		if ( bearing < 0 ) {
			if ( long_diff < 0 ) {
				bearing += 360;
			} else {
				bearing += 180;
			}
		} else
			if (long_diff < 0) 
				bearing += 180;
		   
		return bearing;

#if false  
    # bearing from positions $p1 and $p2 with same datum
    # formulae from "inverse" program (see above)

    set lad1 [lindex $p1 0] ; set lod1 [lindex $p1 1]
    set lad2 [lindex $p2 0] ; set lod2 [lindex $p2 1]
    if { $lad1==$lad2 && $lod1==$lod2 } { return 0 }
    set la1 [expr $lad1*0.01745329251994329576]
    set lo1 [expr $lod1*0.01745329251994329576]
    set la2 [expr $lad2*0.01745329251994329576]
    set lo2 [expr $lod2*0.01745329251994329576]
    if { [expr abs(cos($la1))]<1.e-20 || [expr abs(cos($la2))]<1.e-20 } {
	# use Law of Cosines for Spherical Trigonometry
	# bearing
	set da [expr $la2-$la1] ; set do [expr $lo2-$lo1]
	if { [expr abs($da)] < 1e-20 } {
	    if { [expr abs($do)] < 1e-20 } {
		set b 0
	    } elseif { $do < 0 } {
		set b 270
	    } else { set b 90 }
	} elseif { [expr abs($do)] < 1e-20 } {
	    if { $da < 0 } {
		set b 180
	    } else { set b 0 }
	} else {
	    set b [expr round(atan2(sin($do), \
		                    tan($la2)*cos($la1)-sin($la1)*cos($do)) \
				    *57.29577951308232087684)]
	    if { $b < 0 } {
		if { $do < 0 } { incr b 360 } else { incr b 180 }
	    } elseif { $do < 0 } { incr b 180 }
	}
	return $b
    }
    set dt [EllipsdData $datum]
    set a [lindex $dt 0] ; set f [lindex $dt 1]
    set eps 5e-12
    set r [expr 1-$f]
    set tu1 [expr $r*tan($la1)] ; set tu2 [expr $r*tan($la2)]
    set cu1 [expr 1.0/sqrt($tu1*$tu1+1.0)]
    set su1 [expr $cu1*$tu1]
    set cu2 [expr 1.0/sqrt($tu2*$tu2+1.0)]
    set s [expr $cu1*$cu2]
    set baz [expr $s*$tu2] ; set faz [expr $baz*$tu1]
    set x [expr $lo2-$lo1] ; set d [expr $x+1]
    while { abs($d-$x) > $eps } {
	set sx [expr sin($x)] ; set cx [expr cos($x)]
	set tu1 [expr $cu2*$sx] ; set tu2 [expr $baz-$su1*$cu2*$cx]
	set sy [expr sqrt($tu1*$tu1+$tu2*$tu2)]
	set cy [expr $s*$cx+$faz] ; set y [expr atan2($sy,$cy)]
	set sa [expr $s*$sx/$sy] ; set c2a [expr -$sa*$sa+1.0]
	set cz [expr $faz+$faz]
	if { $cz > 0 } { set cz [expr -$cz/$c2a+$cy] }
	set e [expr $cz*$cz*2-1.0]
	set c [expr ((-3*$c2a+4.0)*$f+4.0)*$c2a*$f/16]
	set d $x
	set x [expr (($e*$cy*$c+$cz)*$sy*$c+$y)*$sa]
	set x [expr (1-$c)*$x*$f+$lo2-$lo1]
    }
    set faz [expr atan2($tu1,$tu2)]
    set b [expr round($faz*57.29577951308232087684)]
    if { $b < 0 } { incr b 360 }
    return $b
#endif


	}

	/**
	   Project a new position from the current position, given bearing
	   and distance.
	*/
	public Position ProjectFrom (double distance, int bearing) {

		Position result = new Position ();

		if (distance * 1000 > Math.PI * a) {
			throw new Exception (String.Format ("Distance too large: {0}", distance));
		}

		return result;

#if false

    # compute coordinates of point at given distance (km) and bearing
    #  (0-360 degrees) from another point
    # algorithm taken from the program "forward" available from
    #  ftp://www.ngs.noaa.gov/pub/pcsoft/for_inv.3d/
    #  It uses the modified Rainsford's Method with Helmert's elliptical
    #  terms, and is effective in any azimuth and at any distance short of
    #  antipodal.
    #  $dist is constrained here to be less than pi*semi-major axis
    global MESS

    foreach "lat0 long0" $pos { break }
    foreach "a f" [EllipsdData $datum] { break }
    if { [set dist [expr $dist*1000]] >= 3.1415926*$a } {
	GMMessage $MESS(distlarge)
	return [list $lat $long]
    }
    set phi [expr $lat0*0.01745329251994329576]
    set lam [expr $long0*0.01745329251994329576]
    set azm [expr $bear*0.01745329251994329576]
    set r [expr 1-$f]
    set tu [expr $r*sin($phi)/cos($phi)]
    set sf [expr sin($azm)] ; set cf [expr cos($azm)]
    set baz 0
    if { [set cf [expr cos($azm)]] != 0 } {
	set baz [expr atan2($tu,$cf)*2]
    }
    set cu [expr 1.0/sqrt($tu*$tu+1)]
    set su [expr $tu*$cu]
    set sa [expr $cu*$sf]
    set c2a [expr -$sa*$sa+1]
    set x [expr sqrt((1.0/$r/$r-1)*$c2a+1)+1]
    set x [expr ($x-2.0)/$x]
    set c [expr 1-$x]
    set c [expr ($x*$x/4.0+1.0)/$c]
    set d [expr (0.375*$x*$x-1)*$x]
    set y [set tu [expr 1.0*$dist/$r/$a/$c]]
    while 1 {
	set sy [expr sin($y)] ; set cy [expr cos($y)]
	set cz [expr cos($baz+$y)]
	set e [expr $cz*$cz*2-1]
	set c $y
	set x [expr $e*$cy]
	set y [expr ((($sy*$sy*4-3)*($e+$e-1)*$cz*$d/6.0+$x)*$d/4.0-$cz)* \
		    $sy*$d+$tu]
	if { abs($y-$c) <= 5e-14 } { break }
    }
    set baz [expr $cu*$cy*$cf-$su*$sy]
    set c [expr $r*sqrt($sa*$sa+$baz*$baz)]
    set d [expr $su*$cy+$cu*$sy*$cf]
    set phi2 [expr atan2($d,$c)]
    set c [expr $cu*$cy-$su*$sy*$cf]
    set x [expr atan2($sy*$sf,$c)]
    set c [expr ((-3*$c2a+4)*$f+4)*$c2a*$f/16.0]
    set d [expr (($e*$cy*$c+$cz)*$sy*$c+$y)*$sa]
    set lam2 [expr $lam+$x-(1-$c)*$d*$f]
    # set baz [expr atan2($sa,$baz)+3.14159265358979323846]

    set lat1 [expr $phi2*57.29577951308232087684]
    set long1 [expr $lam2*57.29577951308232087684]
    return [list $lat1 $long1]
#endif


	}

	/**
	   Utility function to dump position as a string. Useful for debugging.
	*/
	public string AsString () 
	{
		double lo = longitude;
		string ns = "N";
		if (lo < 0) {
			ns = "S";
			lo = - lo;
		}

		double la = latitude;
		string ew = "E";
		if (la < 0) {
			ew = "W";
			la = - la;
		}

		return String.Format ("({2}{0:F5},{3}{1:F5})", lo, la, ns, ew);
	}

}

