using System;
using System.Collections;
using System.IO;
using System.Xml;


/** 
	@file GPXFile 
	@brief Contains GPXFile class
*/

/**
   @class GPXFile
   @brief Class which reads and writes GPX XML files
*/

public class GPXFile {

	string filename;
	Db db;

	public string Filename {
		get {
			return filename;
		}
		set {
			filename = value;
		}
	}

	public GPXFile (Db db, String Filename) {
		this.db = db;
		filename = Filename;
	}


	public void Import () {

		Console.WriteLine ("GPX reading file {0}", filename);

		try {
			XmlDocument doc = new XmlDocument ();
			doc.Load (filename);
			XmlNamespaceManager nsmgr = new XmlNamespaceManager (doc.NameTable);
			nsmgr.AddNamespace (String.Empty, "http://www.topografix.com/GPX/1/0");
			nsmgr.AddNamespace ("__defns", "http://www.topografix.com/GPX/1/0");

			ImportWaypoints (doc, nsmgr);
			ImportRoutes (doc, nsmgr);
			ImportTracks (doc, nsmgr);

		} catch (Exception e) {
			// FIXME: handle error more gracefully
			throw new System.Exception ("Error while importing GPX file", e);
		}

	}


	/** Import a waypoint from the information at the given	node.
		@param node The XML node which holds the waypoint information.
		@param nsmgr The XML namespace manager to be used.
	*/
	private Waypoint ImportWaypoint (XmlNode wp_node, XmlNamespaceManager nsmgr) 
	{
		//Console.WriteLine ("ImportWaypoint XML dump: {0}", wp_node.OuterXml);

		// Create new position from XML attributes
		Position position = new Position ();
		string s = wp_node.Attributes.GetNamedItem ("lat").InnerText;
		position.Latitude = (s as IConvertible).ToDouble (null);
		s = wp_node.Attributes.GetNamedItem ("lon").InnerText;
		position.Longitude = (s as IConvertible).ToDouble (null);

		// Create a new waypoint at this position
		Waypoint waypoint = db.Waypoints.CreateWaypoint (position);

		// Set all other attributes of the waypoint from XML children
		XmlNode node = wp_node.SelectSingleNode ("__defns:ele", nsmgr);
		if (node == null) {
			// FIXME: 0 is not the same as unknown...
			waypoint.Elevation = 0;
		} else {
			s = node.InnerText;
			waypoint.Elevation = (s as IConvertible).ToDouble (null);
		}

		node = wp_node.SelectSingleNode ("__defns:name", nsmgr);
		if (node == null) {
			// FIXME: should provide good default, unique name
			waypoint.Name = "No name";
		} else {
			waypoint.Name = node.InnerText;
		}

		node = wp_node.SelectSingleNode ("__defns:cmt", nsmgr);
		if (node == null) {
			waypoint.Comment = "";
		} else {
			waypoint.Comment = node.InnerText;
		}

		node = wp_node.SelectSingleNode ("__defns:desc", nsmgr);
		if (node == null) {
				waypoint.Description = "";
		} else {
				waypoint.Description = node.InnerText;
		}
				
		// FIXME: still missing a few elements that should be supported, in particular time and sym (and perhaps type)

		Console.WriteLine (waypoint.AsString ());

		// Commit these changes into the store.
		db.Waypoints.Commit (waypoint);

		return waypoint;
	}



	/** Import the waypoints from the GPX file
		@param doc The parsed XML document
		@param nsmger Namespace info to properly parse the document
	*/
	private void ImportWaypoints (XmlDocument doc, XmlNamespaceManager nsmgr)
	{
		XmlNodeList waypoints = doc.GetElementsByTagName ("wpt");
		IEnumerator iter = waypoints.GetEnumerator ();
		iter.Reset ();
		while (iter.MoveNext ()) {
			ImportWaypoint ( (XmlNode) iter.Current, nsmgr);
		}
	}


	/** Import the routes from the GPX file
		@param doc The parsed XML document
		@param nsmger Namespace info to properly parse the document
	*/
	private void ImportRoutes (XmlDocument doc, XmlNamespaceManager nsmgr)
	{
		XmlNodeList routes = doc.GetElementsByTagName ("rte");
		IEnumerator iter = routes.GetEnumerator ();
		iter.Reset ();
		while (iter.MoveNext ()) {
			XmlNode route_node = (XmlNode) iter.Current;
			Console.WriteLine ("Route XML dump: {0}", route_node.OuterXml);
			
			/* 
			   Get the attributes of the route and create the route
			   itself.
			*/
			string name = route_node.SelectSingleNode ("__defns:name", nsmgr).InnerText;
			string description = route_node.SelectSingleNode ("__defns:desc", nsmgr).InnerText;
			Route route = db.Routes.CreateRoute (name, description);
			
			
			/* 
			   Each route contains a number of ordered route
			   points. For this route we need to cycle through the
			   route points and string them together.
			*/
			XmlNodeList route_points = route_node.SelectNodes ("__defns:rtept", nsmgr);
			IEnumerator rp_iter = route_points.GetEnumerator ();
			rp_iter.Reset ();
			while (rp_iter.MoveNext ()) {
				Waypoint waypoint = ImportWaypoint ( (XmlNode) rp_iter.Current, nsmgr);
				route.WaypointAdd (waypoint);
			}

			/* Commit all changes, including added waypoints */
			db.Routes.Commit (route);
		}
	}


	/** Import the tracks from the GPX file.
		@param doc The parsed XML document
		@param nsmger Namespace info to properly parse the document
	*/
	private void ImportTracks (XmlDocument doc, XmlNamespaceManager nsmgr)
	{
#if false
		XmlNodeList routes = doc.GetElementsByTagName ("trk");
		IEnumerator iter = routes.GetEnumerator ();
		iter.Reset ();
		while (iter.MoveNext ()) {
			XmlNode route_node = (XmlNode) iter.Current;
			Console.WriteLine ("Route XML dump: {0}", route_node.OuterXml);
			
			/* 
			   Get the attributes of the route and create the route
			   itself.
			*/
			string name = route_node.SelectSingleNode ("__defns:name", nsmgr).InnerText;
			string description = route_node.SelectSingleNode ("__defns:desc", nsmgr).InnerText;
			Route route = db.Routes.CreateRoute (name, description);
			
			
			/* 
			   Each route contains a number of ordered route
			   points. For this route we need to cycle through the
			   route points and string them together.
			*/
			XmlNodeList route_points = route_node.SelectNodes ("__defns:rtept", nsmgr);
			IEnumerator rp_iter = route_points.GetEnumerator ();
			rp_iter.Reset ();
			while (rp_iter.MoveNext ()) {
				Waypoint waypoint = ImportWaypoint ( (XmlNode) rp_iter.Current, nsmgr);
				route.WaypointAdd (waypoint);
			}

			/* Commit all changes, including added waypoints */
			db.Routes.Commit (route);
		}
#else
		/**
		 * @todo Import track information from the GPX file. This also
		   requires a decision on how to handle track points. They are
		   very much like waypoints, but their use is quite different,
		   and using them as normal waypoints may clutter up lists of
		   waypoints in the application.
		*/
		Console.WriteLine ("TODO: track import");
#endif

	}
	
}
