using Gdk;
using GLib;
using Gnome;
using Gtk;
using GtkSharp;
using System;

/// \file MapView.cs
/// \brief The View that handles displaying and interacting with the map.

/// \class MapView
/// \brief The class that manages the View on the Map.

public class MapView : EventBox {

	/** 
		position is the top left location of the current mapview.
	*/
	Position base_position;

	public static double ScaleMultiplier = 1.3;
	public static double DefaultPixelsPerUnit = 100;

	Db db;
	Canvas canvas;
	CanvasGroup root;

	CanvasWaypoint [] cw_list = new CanvasWaypoint[9];

	/**
	   Handle a motion notify event on the map view.  We handle this
	   here because only the map view knows about things like
	   underlying coordinates.
	*/
	public void HandleMotionNotifyEvent (object sender, Gtk.MotionNotifyEventArgs args, Gtk.Label status)
	{
		double wx, wy;

		canvas.C2w ((int) args.Event.X, (int) args.Event.Y, out wx, out wy);
		root.W2i (ref wx, ref wy);

		Position position = new Position (wx, wy);
		status.Text = position.AsString ();
	}


	/** 
		Zoom in on the canvas.
	*/
	public void HandleZoomInCommand () 
	{
		/// @todo check on maximum zoom factor
		canvas.PixelsPerUnit *= ScaleMultiplier;
	}

	/**
	   Zoom out on the canvas.
	*/
	public void HandleZoomOutCommand () 
	{
		/// @todo check on maximum zoom factor
		canvas.PixelsPerUnit /= ScaleMultiplier;
	}

	/** 
		Reset zoom to the default value
	*/
	public void HandleZoomDefaultCommand ()
	{
		canvas.PixelsPerUnit = DefaultPixelsPerUnit;
	}


	/**
	   Scroll to the given location on the screen.
	   @param latitude The latitude to scroll to
	   @param longitude The longitude to scroll to
	*/
	public void ScrollTo (double latitude, double longitude)
	{
		// Convert the longitude and latitude to world coordinates
		root.I2w (ref latitude, ref longitude);
		// Convert the world coordinates to canvas coordinates
		int xc, yc;
		canvas.W2c (latitude, longitude, out xc, out yc);

		// Move canvas there
		canvas.ScrollTo (xc - (int) (canvas.Allocation.Width / 2), yc - (int) (canvas.Allocation.Height / 2));
	}

	/**
	   Update the map view. This needs to be sophisticated, but for
	   now just scan the database and draw everything.
	*/

	public void Update () 
	{
		Waypoint [] waypoints = db.Waypoints.Query ();

		int i = 0;

		Console.WriteLine ("Updating mapview with following waypoints:");
		foreach (Waypoint waypoint in waypoints) {
			Console.WriteLine ("Draw waypoint {0}", waypoint.AsString ());
			cw_list[i] = new CanvasWaypoint (root, waypoint);
			i++;
		}

	}




	public MapView (Db db, Gnome.Canvas canvas)
		: base ()
	{
		this.db = db;
		this.canvas = canvas;
		this.root = canvas.Root ();

		// Map the earth coordinates (item coordinates) onto the
		// canvas world coordinates
		double [] affinematrix = {1, 0, 0, -1, 180, 180};
		root.AffineAbsolute (affinematrix);

		// Set the scrolling region to the whole world
		canvas.SetScrollRegion (0, 0, 360, 360);
		canvas.PixelsPerUnit = DefaultPixelsPerUnit;
		ScrollTo (4.33320, 52.01001);

		// Update the display (ie. draw the waypoints)
		Update ();

	}
}

