using Gdk;
using Gnome;
using Gtk;
using Mono.Data.SqliteClient;
using System.Collections;
using System.IO;
using System;

/**
   @file RouteStore.cs
   @brief Defines the Route and RouteStore classes.
*/


/**
   @class Route
   @brief A route in GPS terms as can be uploaded to a GPS receiver.
   A route is an ordered collection of waypoints with some added data.
*/

public class Route : DbItem, IComparable {

	private string name;
	/// The name of the waypoint as it is also known on the GPS device
	public string Name {
		set { name = value;	}
		get { return name; }
	}
	
	private string description;
	///	A more verbose explanation that will not be downloaded to the GPS.
	public string Description {
		set { description = value; }
		get { return description;  }
	}
			
	private ArrayList route_points = new ArrayList ();
	/** The ordered list of route points. Each element in this list is
		a normal Waypoint.
	*/
	public ArrayList RoutePoints {
		get { return route_points; }
	}
	

	// You are not supposed to invoke these constructors outside of the
	// associated ...Store class.
	public Route (uint id, string name, string description)
		: base (id)
	{
		this.name = name;
		this.description = description;
	}


	// IComparer.
	public int CompareTo (object obj)
	{
		Route route = obj as Route;

		return Name.CompareTo (route.Name);
	}

	/**
	   Utility function to dump the route as a string. Useful for
	   debugging.
	*/
	public string AsString ()
	{
		string result = String.Format ("{0} ({1}): ", name, description);
		
		foreach (Waypoint waypoint in route_points) {
			result += waypoint.AsString ();
		}

		return result;
	}

	/** Add a waypoint to this route. The added waypoint will be the
		last waypoint.
		@param waypoint The waypoint to be added to this route.
	*/
	public void WaypointAdd (Waypoint waypoint)
	{
		route_points.Add (waypoint);
	}
		

}

public delegate void RouteCreatedHandler(Route w);
public delegate void RouteChangedHandler(Route w);
public delegate void RouteDeletedHandler(Route w);

/**
   @class RouteStore
   @brief The storage facility for routes.
*/

public class RouteStore : DbStore {

	public event RouteCreatedHandler RouteCreated;
	public event RouteChangedHandler RouteChanged;
	public event RouteDeletedHandler RouteDeleted;
	

	/** Create the RouteStorage table */
	private void CreateTable ()
	{
		Console.WriteLine ("Creating routes table");

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText =
			"CREATE TABLE routes (                      " +
			"	id            INTEGER PRIMARY KEY NOT NULL," +
			"   name          TEXT NOT NULL,               " +
			"   description   TEXT                         " +
			")";

		command.ExecuteNonQuery ();
		command.Dispose ();
	}

	/** Initialize the RouteStore with the sqlite connnection
		@param connection Sqlite connection
		@param is_new Whether the database needs to be created
	*/
		
	public RouteStore (SqliteConnection connection, bool is_new)
		: base (connection, true)
	{
		if (is_new) {
			CreateTable ();
		}
	}

	/** Create a new route as part of the route store.
		Use default values for all parameters not given.
		@param name The name of the route
		@param description The description of the route.
	*/
	public Route CreateRoute (string name, string description) 
	{
		Console.WriteLine ("Creating new route {0} ({1}", name, description);

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format
			("INSERT INTO routes (name, description) " +
			 "VALUES ('{0}', '{1}')",
			 SqlString (name),
			 SqlString (description)
			 );

		command.ExecuteScalar ();
		command.Dispose ();

		uint id = (uint) Connection.LastInsertRowId;

		Route route = new Route (id, name, description);
		AddToCache (route);
		if (RouteCreated != null)
			RouteCreated (route);	
		return route;
	}

	/** Look up a route in the cache.
		@param id The id of the route.
	*/
	public override DbItem Get (uint id)
	{
		return LookupInCache (id) as Route;
	}
	
	/** Remove a route from the store.
		@param item The item (route) to be removed.
	*/
	public override void Remove (DbItem item)
	{
		RemoveFromCache (item);
		
		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format ("DELETE FROM routes WHERE id = {0}", item.Id);
		command.ExecuteNonQuery ();

		command.Dispose ();
		RouteDeleted((Route)item);
	}

	/** Commit changes to an already existing route to the store.
		@param item The item (route) to be committed.
	*/
	public override void Commit (DbItem item)
	{
		Route route = item as Route;

		Console.WriteLine ("Committing data for route {0}", route.Name);

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format ("UPDATE routes SET          " +
						     "    name = '{0}', " +
							 "    description = '{1}' "+
						     "WHERE id = {2}           ",
											 SqlString (route.Name),
											 SqlString (route.Description),
											 route.Id);
		command.ExecuteNonQuery ();
		
		/**
		   @todo Also commit the waypoints array. Needs another
		   database table to store relation between waypoints and
		   routes. Probably also needs additional code in
		   WaypointStore to check if a waypoint still belongs to a
		   particular route when deleting it.
		*/
		Console.WriteLine ("TODO: also commit the waypoints array!");

		command.Dispose ();
	}

	/** 
		Run a query on all current routes. This query returns all
		stored routes.
	*/
	public Route [] Query ()
	{
		string query_string = 
			"SELECT routes.id,   " +
			" routes.name,       " +
			" routes.description " +
			" FROM routes";
		return Query (query_string);
	}

	/** 
		Run the actual query as defined in the argument. The argument
		must be valid SQL code.
		@param query A string with a query as valid SQL code.
	*/

	public Route [] Query (string query) 
	{
		Console.WriteLine ("Query Start {0}", System.DateTime.Now.ToLongTimeString() );

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;
		command.CommandText = query;
		SqliteDataReader reader = command.ExecuteReader ();

		Console.WriteLine ("Query Mid {0}", System.DateTime.Now.ToLongTimeString() );

		ArrayList id_list = new ArrayList ();
		while (reader.Read ()) {
			uint id = Convert.ToUInt32 (reader [0]);
			Route route = LookupInCache (id) as Route;

			if (route == null) {
				Console.WriteLine ("Route {0} not found in cache", id);
				route = new Route (id, 
								   reader [1].ToString (), 
								   reader [2].ToString ());
			}

			Console.WriteLine ("TODO: also read the waypoints and build the arraylist!");

			id_list.Add (route);
		}

		Console.WriteLine ("Query End {0}", System.DateTime.Now.ToLongTimeString() );
		
		command.Dispose ();

		return id_list.ToArray (typeof (Route)) as Route [];
	}



#if false
	static void Main (string [] args)
	{
		Program program = new Program ("TagStoreTest", "0.0", Modules.UI, args);

		const string path = "/tmp/TagStoreTest.db";

		try {
			File.Delete (path);
		} catch {}

		Db db = new Db (path, true);

		Category people_category = db.Tags.CreateCategory (null, "People");
		Tag anna_tag = db.Tags.CreateTag (people_category, "Anna");
		Tag ettore_tag = db.Tags.CreateTag (people_category, "Ettore");
		Tag miggy_tag = db.Tags.CreateTag (people_category, "Miggy");
		miggy_tag.SortPriority = -1;
		db.Tags.Commit (miggy_tag);

		Category places_category = db.Tags.CreateCategory (null, "Places");
		Tag milan_tag = db.Tags.CreateTag (places_category, "Milan");
		Tag boston_tag = db.Tags.CreateTag (places_category, "Boston");

		Category exotic_category = db.Tags.CreateCategory (places_category, "Exotic");
		Tag bengalore_tag = db.Tags.CreateTag (exotic_category, "Bengalore");
		Tag manila_tag = db.Tags.CreateTag (exotic_category, "Manila");
		Tag tokyo_tag = db.Tags.CreateTag (exotic_category, "Tokyo");

		tokyo_tag.Category = places_category;
		tokyo_tag.Name = "Paris";
		db.Tags.Commit (tokyo_tag);

		db.Dispose ();

		db = new Db (path, false);
		Dump (db.Tags.RootCategory, 0);
	}
#endif

}
