using Gdk;
using Gnome;
using Gtk;
using Mono.Data.SqliteClient;
using System.Collections;
using System.IO;
using System;

/**
   @file TrackStore.cs
   @brief Defines the Track and TrackStore classes.
*/


/**
   @class Track
   @brief A track in GPS terms as can be uploaded to a GPS receiver.
   A track is an ordered collection of waypoints with some added data.
*/

public class Track : DbItem, IComparable {

	private string name;
	/// The name of the waypoint as it is also known on the GPS device
	public string Name {
		set { name = value;	}
		get { return name; }
	}
	
	private string description;
	///	A more verbose explanation that will not be downloaded to the GPS.
	public string Description {
		set { description = value; }
		get { return description;  }
	}
			
	private ArrayList track_points = new ArrayList ();
	/** The ordered list of track points. Each element in this list is
		a normal Waypoint.
	*/
	public ArrayList TrackPoints {
		get { return track_points; }
	}
	

	// You are not supposed to invoke these constructors outside of the
	// associated ...Store class.
	public Track (uint id, string name, string description)
		: base (id)
	{
		this.name = name;
		this.description = description;
	}


	// IComparer.
	public int CompareTo (object obj)
	{
		Track track = obj as Track;

		return Name.CompareTo (track.Name);
	}

	/**
	   Utility function to dump the track as a string. Useful for
	   debugging.
	*/
	public string AsString ()
	{
		string result = String.Format ("{0} ({1}): ", name, description);
		
		foreach (Waypoint waypoint in track_points) {
			result += waypoint.AsString ();
		}

		return result;
	}

	/** Add a waypoint to this track. The added waypoint will be the
		last waypoint.
		@param waypoint The waypoint to be added to this track.
	*/
	public void WaypointAdd (Waypoint waypoint)
	{
		track_points.Add (waypoint);
	}
		

}

public delegate void TrackCreatedHandler(Track w);
public delegate void TrackChangedHandler(Track w);
public delegate void TrackDeletedHandler(Track w);

/**
   @class TrackStore
   @brief The storage facility for tracks.
*/

public class TrackStore : DbStore {

	public event TrackCreatedHandler TrackCreated;
	public event TrackChangedHandler TrackChanged;
	public event TrackDeletedHandler TrackDeleted;
	

	/** Create the TrackStorage table */
	private void CreateTable ()
	{
		Console.WriteLine ("Creating tracks table");

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText =
			"CREATE TABLE tracks (                      " +
			"	id            INTEGER PRIMARY KEY NOT NULL," +
			"   name          TEXT NOT NULL,               " +
			"   description   TEXT                         " +
			")";

		command.ExecuteNonQuery ();
		command.Dispose ();
	}

	/** Initialize the TrackStore with the sqlite connnection
		@param connection Sqlite connection
		@param is_new Whether the database needs to be created
	*/
		
	public TrackStore (SqliteConnection connection, bool is_new)
		: base (connection, true)
	{
		if (is_new) {
			CreateTable ();
		}
	}

	/** Create a new track as part of the track store.
		Use default values for all parameters not given.
		@param name The name of the track
		@param description The description of the track.
	*/
	public Track CreateTrack (string name, string description) 
	{
		Console.WriteLine ("Creating new track {0} ({1}", name, description);

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format
			("INSERT INTO tracks (name, description) " +
			 "VALUES ('{0}', '{1}')",
			 SqlString (name),
			 SqlString (description)
			 );

		command.ExecuteScalar ();
		command.Dispose ();

		uint id = (uint) Connection.LastInsertRowId;

		Track track = new Track (id, name, description);
		AddToCache (track);
		if (TrackCreated != null)
			TrackCreated (track);	
		return track;
	}

	/** Look up a track in the cache.
		@param id The id of the track.
	*/
	public override DbItem Get (uint id)
	{
		return LookupInCache (id) as Track;
	}
	
	/** Remove a track from the store.
		@param item The item (track) to be removed.
	*/
	public override void Remove (DbItem item)
	{
		RemoveFromCache (item);
		
		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format ("DELETE FROM tracks WHERE id = {0}", item.Id);
		command.ExecuteNonQuery ();

		command.Dispose ();
		TrackDeleted((Track)item);
	}

	/** Commit changes to an already existing track to the store.
		@param item The item (track) to be committed.
	*/
	public override void Commit (DbItem item)
	{
		Track track = item as Track;

		Console.WriteLine ("Committing data for track {0}", track.Name);

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format ("UPDATE tracks SET          " +
						     "    name = '{0}', " +
							 "    description = '{1}' "+
						     "WHERE id = {2}           ",
											 SqlString (track.Name),
											 SqlString (track.Description),
											 track.Id);
		command.ExecuteNonQuery ();

		Console.WriteLine ("TODO: also commit the waypoints array!");

		command.Dispose ();
	}

	/** 
		Run a query on all current tracks. This query returns all
		stored tracks.
	*/
	public Track [] Query ()
	{
		string query_string = 
			"SELECT tracks.id,   " +
			" tracks.name,       " +
			" tracks.description " +
			" FROM tracks";
		return Query (query_string);
	}

	/** 
		Run the actual query as defined in the argument. The argument
		must be valid SQL code.
		@param query A string with a query as valid SQL code.
	*/

	public Track [] Query (string query) 
	{
		Console.WriteLine ("Query Start {0}", System.DateTime.Now.ToLongTimeString() );

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;
		command.CommandText = query;
		SqliteDataReader reader = command.ExecuteReader ();

		Console.WriteLine ("Query Mid {0}", System.DateTime.Now.ToLongTimeString() );

		ArrayList id_list = new ArrayList ();
		while (reader.Read ()) {
			uint id = Convert.ToUInt32 (reader [0]);
			Track track = LookupInCache (id) as Track;

			if (track == null) {
				Console.WriteLine ("Track {0} not found in cache", id);
				track = new Track (id, 
								   reader [1].ToString (), 
								   reader [2].ToString ());
			}

			Console.WriteLine ("TODO: also read the waypoints and build the arraylist!");

			id_list.Add (track);
		}

		Console.WriteLine ("Query End {0}", System.DateTime.Now.ToLongTimeString() );
		
		command.Dispose ();

		return id_list.ToArray (typeof (Track)) as Track [];
	}



#if false
	static void Main (string [] args)
	{
		Program program = new Program ("TagStoreTest", "0.0", Modules.UI, args);

		const string path = "/tmp/TagStoreTest.db";

		try {
			File.Delete (path);
		} catch {}

		Db db = new Db (path, true);

		Category people_category = db.Tags.CreateCategory (null, "People");
		Tag anna_tag = db.Tags.CreateTag (people_category, "Anna");
		Tag ettore_tag = db.Tags.CreateTag (people_category, "Ettore");
		Tag miggy_tag = db.Tags.CreateTag (people_category, "Miggy");
		miggy_tag.SortPriority = -1;
		db.Tags.Commit (miggy_tag);

		Category places_category = db.Tags.CreateCategory (null, "Places");
		Tag milan_tag = db.Tags.CreateTag (places_category, "Milan");
		Tag boston_tag = db.Tags.CreateTag (places_category, "Boston");

		Category exotic_category = db.Tags.CreateCategory (places_category, "Exotic");
		Tag bengalore_tag = db.Tags.CreateTag (exotic_category, "Bengalore");
		Tag manila_tag = db.Tags.CreateTag (exotic_category, "Manila");
		Tag tokyo_tag = db.Tags.CreateTag (exotic_category, "Tokyo");

		tokyo_tag.Category = places_category;
		tokyo_tag.Name = "Paris";
		db.Tags.Commit (tokyo_tag);

		db.Dispose ();

		db = new Db (path, false);
		Dump (db.Tags.RootCategory, 0);
	}
#endif

}
