using Gdk;
using Gnome;
using Gtk;
using Mono.Data.SqliteClient;
using System.Collections;
using System.IO;
using System;

/**
   @file WaypointStore.cs
   @brief Defines the Waypoint and WaypointStore classes.
*/


/**
   @class Waypoint
   @brief A waypoint in GPS terms as can be uploaded to a GPS receiver.
*/

public class Waypoint : DbItem, IComparable {
	private Position position;
	/// The location of the waypoint as defined by a GPS Position.
	public Position Position {
		set { position = value;	}
		get { return position;	}
	}

	private double elevation;
	/// The elevation of the waypoint at the given position, in meters.
	public double Elevation {
		set { elevation = value; }
		get { return elevation;  }
	}

	private string name;
	/// The name of the waypoint as it is also known on the GPS device
	public string Name {
		set { name = value;	}
		get { return name; }
	}
	
	private string comment;
	/// An additional comment for the waypoint, to be downloaded to the GPS.
	public string Comment {
		set { comment = value; }
		get { return comment;  }
	}

	private string description;
	///	A more verbose explanation that will not be downloaded to the GPS.
	public string Description {
		set { description = value; }
		get { return description;  }
	}
			
	private DateTime time;
	/// The time of creation of the waypoint.
	public DateTime Time {
		set { time = value;	}
		get { return time;	}
	}
	

	// You are not supposed to invoke these constructors outside of the
	// associated ...Store class.
	public Waypoint (uint id, Position position)
		: base (id)
	{
		this.position = position;
	}

	public Waypoint (uint id, Position position, double elevation, string name, string comment, string description)
		: this (id, position)
	{
		this.elevation = elevation;
		this.name = name;
		this.comment = comment;
		this.description = description;
	}


	// IComparer.
	public int CompareTo (object obj)
	{
		Waypoint waypoint = obj as Waypoint;

		return Name.CompareTo (waypoint.Name);
	}

	/**
	   Utility function to dump the waypoint as a string. Useful for
	   debugging.
	*/
	public string AsString ()
	{
		return String.Format ("{2} ({3}): {0} at {1} meters ({4})", position.AsString (), elevation, name, comment, description);
	}

}

public delegate void WaypointCreatedHandler(Waypoint w);
public delegate void WaypointChangedHandler(Waypoint w);
public delegate void WaypointDeletedHandler(Waypoint w);

/**
   @class WaypointStore
   @brief The storage facility for waypoints.
*/

public class WaypointStore : DbStore {

	public event WaypointCreatedHandler WaypointCreated;
	public event WaypointChangedHandler WaypointChanged;
	public event WaypointDeletedHandler WaypointDeleted;
	

	/** Create the WaypointStorage table */
	private void CreateTable ()
	{
		Console.WriteLine ("Creating waypoints table");

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText =
			"CREATE TABLE waypoints (                      " +
			"	id            INTEGER PRIMARY KEY NOT NULL," +
			"   latitude      REAL,                        " +
			"   longitude     REAL,		     	           " +
			"   elevation     REAL,                        " +
			"   name          TEXT NOT NULL,               " +
			"   comment       TEXT,                        " +
			"   description   TEXT                         " +
			")";

		command.ExecuteNonQuery ();
		command.Dispose ();
	}

	/** Initialize the WaypointStore with the sqlite connnection
		@param connection Sqlite connection
		@param is_new Whether the database needs to be created
	*/
		
	public WaypointStore (SqliteConnection connection, bool is_new)
		: base (connection, true)
	{
		if (is_new) {
			CreateTable ();
		}
	}

	/** Create a new waypoint as part of the waypoint store.
		Use default values for all paramaters not given.
		@param position The Position at which the waypoint is located.
	*/
	public Waypoint CreateWaypoint (Position position) 
	{
		return CreateWaypoint (position, 0, "NO NAME", "", "");
	}

	/** Create a new waypoint as part of the waypoint store. Give all
		values of a waypoint explicitly.
		@param position The Position at which the waypoint is located.
		@param elevation The elevation of the waypoint (in meters).
		@param name The GPS name of the waypoint (mandatory).
		@param comment The GPS comment of the waypoint.
		@param description Annotation of the waypoint.
	*/
	public Waypoint CreateWaypoint (Position position, double elevation, string name, string comment, string description)
	{
		Console.WriteLine ("Creating new waypoint {0}", name);

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format
			("INSERT INTO waypoints (latitude, longitude, elevation, name, comment, description) " +
			 "VALUES ({0}, {1}, {2}, '{3}', '{4}', '{5}')",
			 position.Latitude,
			 position.Longitude,
			 elevation,
			 SqlString (name),
			 SqlString (comment),
			 SqlString (description)
			 );

		command.ExecuteScalar ();
		command.Dispose ();

		uint id = (uint) Connection.LastInsertRowId;

		Waypoint waypoint = new Waypoint (id, position, elevation, name, comment, description);
		AddToCache (waypoint);
		if (WaypointCreated != null)
			WaypointCreated (waypoint);	
		return waypoint;
	}

	/** Look up a waypoint in the cache.
		@param id The id of the waypoint.
	*/
	public override DbItem Get (uint id)
	{
		return LookupInCache (id) as Waypoint;
	}
	
	/** Remove a waypoint from the store.
		@param item The item (waypoint) to be removed.
	*/
	public override void Remove (DbItem item)
	{
		RemoveFromCache (item);
		
		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format ("DELETE FROM waypoints WHERE id = {0}", item.Id);
		command.ExecuteNonQuery ();

		command.Dispose ();
		WaypointDeleted((Waypoint)item);
	}

	/** Commit changes to an already existing waypoint to the store.
		@param item The item (waypoint) to be committed.
	*/
	public override void Commit (DbItem item)
	{
		Waypoint waypoint = item as Waypoint;

		Console.WriteLine ("Committing data for waypoint {0}", waypoint.Name);

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;

		command.CommandText = String.Format ("UPDATE waypoints SET          " +
						     "    latitude = '{0}',        " +
						     "    longitude = '{1}',   " +
						     "    elevation = '{2}',   " +
						     "    name = '{3}', " +
						     "    comment = '{4}',	       " +
							 "    description = '{5}' "+
						     "WHERE id = {6}           ",
											 waypoint.Position.Latitude,
											 waypoint.Position.Longitude,
											 waypoint.Elevation,
											 SqlString (waypoint.Name),
											 SqlString (waypoint.Comment),
											 SqlString (waypoint.Description),
											 waypoint.Id);
		command.ExecuteNonQuery ();

		command.Dispose ();
	}

	/** 
		Run a query on all current waypoints. This query returns all
		stored waypoints.
	*/
	public Waypoint [] Query ()
	{
		string query_string = 
			"SELECT waypoints.id,   " +
			" waypoints.latitude,   " +
			" waypoints.longitude,  " +
			" waypoints.elevation,  " +
			" waypoints.name,       " +
			" waypoints.comment,    " +
			" waypoints.description " +
			" FROM waypoints";
		return Query (query_string);
	}

	/** 
		Run the actual query as defined in the argument. The argument
		must be valid SQL code.
		@param query A string with a query as valid SQL code.
	*/

	public Waypoint [] Query (string query) 
	{
		Console.WriteLine ("Query Start {0}", System.DateTime.Now.ToLongTimeString() );

		SqliteCommand command = new SqliteCommand ();
		command.Connection = Connection;
		command.CommandText = query;
		SqliteDataReader reader = command.ExecuteReader ();

		Console.WriteLine ("Query Mid {0}", System.DateTime.Now.ToLongTimeString() );

		ArrayList id_list = new ArrayList ();
		while (reader.Read ()) {
			uint id = Convert.ToUInt32 (reader [0]);
			Waypoint waypoint = LookupInCache (id) as Waypoint;

			if (waypoint == null) {
				Console.WriteLine ("Waypoint {0} not found in cache", id);
				Position position = new Position (Convert.ToDouble (reader [1]), Convert.ToDouble(reader [2]));
				waypoint = new Waypoint (id, position, 
										 Convert.ToDouble( reader [3]), 
										 reader [4].ToString (), 
										 reader [5].ToString (), 
										 reader [6].ToString ());
			}

			id_list.Add (waypoint);
		}

		Console.WriteLine ("Query End {0}", System.DateTime.Now.ToLongTimeString() );
		
		command.Dispose ();

		return id_list.ToArray (typeof (Waypoint)) as Waypoint [];
	}



#if false
	static void Main (string [] args)
	{
		Program program = new Program ("TagStoreTest", "0.0", Modules.UI, args);

		const string path = "/tmp/TagStoreTest.db";

		try {
			File.Delete (path);
		} catch {}

		Db db = new Db (path, true);

		Category people_category = db.Tags.CreateCategory (null, "People");
		Tag anna_tag = db.Tags.CreateTag (people_category, "Anna");
		Tag ettore_tag = db.Tags.CreateTag (people_category, "Ettore");
		Tag miggy_tag = db.Tags.CreateTag (people_category, "Miggy");
		miggy_tag.SortPriority = -1;
		db.Tags.Commit (miggy_tag);

		Category places_category = db.Tags.CreateCategory (null, "Places");
		Tag milan_tag = db.Tags.CreateTag (places_category, "Milan");
		Tag boston_tag = db.Tags.CreateTag (places_category, "Boston");

		Category exotic_category = db.Tags.CreateCategory (places_category, "Exotic");
		Tag bengalore_tag = db.Tags.CreateTag (exotic_category, "Bengalore");
		Tag manila_tag = db.Tags.CreateTag (exotic_category, "Manila");
		Tag tokyo_tag = db.Tags.CreateTag (exotic_category, "Tokyo");

		tokyo_tag.Category = places_category;
		tokyo_tag.Name = "Paris";
		db.Tags.Commit (tokyo_tag);

		db.Dispose ();

		db = new Db (path, false);
		Dump (db.Tags.RootCategory, 0);
	}
#endif

}
